"""
    conformal_cubed_sphere_mapping(x, y)

Conformal mapping of a cube onto a sphere. Maps `(x, y)` on the north-pole face of a cube
to `(X, Y, Z)` coordinates in physical space. The face is oriented normal to `Z`-axis with
`X` and `Y` increasing with `x` and `y`.

The input coordinates must lie within the range ``-1 ≤ x ≤ 1``,  ``-1 ≤ y ≤ 1``.

This numerical conformal mapping is described by Rančić et al. (1996).

This is a Julia translation of [MATLAB code from MITgcm](http://wwwcvs.mitgcm.org/viewvc/MITgcm/MITgcm_contrib/high_res_cube/matlab-grid-generator/map_xy2xyz.m?view=markup) that is based on
Fortran 77 code from Jim Purser & Misha Rančić.

References
==========

- Rančić et al., (1996). A global shallow-water model using an expanded spherical cube - Gnomonic versus conformal
  coordinates, _Quarterly Journal of the Royal Meteorological Society_.
"""
function conformal_cubed_sphere_mapping(x, y)
    X = xᶜ = abs(x)
    Y = yᶜ = abs(y)

    kxy = yᶜ > xᶜ

    xᶜ = 1 - xᶜ
    yᶜ = 1 - yᶜ

    kxy && (xᶜ = 1 - Y)
    kxy && (yᶜ = 1 - X)

    Z = ((xᶜ + im * yᶜ) / 2)^4
    W = W_Rancic(Z)

    im³ = im^(1/3)
    ra = √3 - 1
    cb = -1 + im
    cc = ra * cb / 2

    W = im³ * (W * im)^(1/3)
    W = (W - ra) / (cb + cc * W)
    X, Y = reim(W)

    H = 2 / (1 + X^2 + Y^2)
    X = X * H
    Y = Y * H
    Z = H - 1

    if kxy
        X, Y = Y, X
    end

    y < 0 && (Y = -Y)
    x < 0 && (X = -X)

    # Fix truncation for x = 0 or y = 0.
    x == 0 && (X = 0)
    y == 0 && (Y = 0)

    return X, Y, Z
end

"""
    conformal_cubed_sphere_inverse_mapping(X, Y, Z)

Inverse mapping for conformal cube sphere for quadrant of north-pole face in which `X` and `Y` are both
positive. All other mappings to other cube face coordinates can be recovered from rotations of this map.
There a 3 other quadrants for the north-pole face and five other faces for a total of twenty-four quadrants.
Because of symmetry only the reverse for a single quadrant is needed. Because of branch cuts and the complex
transform the inverse mappings are multi-valued in general, using a single quadrant case allows a simple
set of rules to be applied.

The mapping is valid for the cube face quadrant defined by ``0 < x < 1`` and ``0 < y < 1``, where a full cube
face has extent ``-1 < x < 1`` and ``-1 < y < 1``. The quadrant for the mapping is from a cube face that has
"north-pole" at its center ``(x=0, y=0)``. i.e., has `X, Y, Z = (0, 0, 1)` at its center. The valid ranges of `X`
and `Y` for this mapping and convention are a quadrant defined be geodesics that connect the points A, B, C and D,
on the shell of a sphere of radius ``R`` with `X`, `Y` coordinates as follows

```
A = (0, 0)
B = (√2, 0)
C = (√3/3, √3/3)
D = (0, √2)
```
"""
function conformal_cubed_sphere_inverse_mapping(X, Y, Z)
    H  = Z + 1
    Xˢ = X / H
    Yˢ = Y / H
    ω  = Xˢ + im * Yˢ

    ra = √3 - 1
    cb = -1 + im
    cc = ra * cb / 2
    ω⁰ = (ω * cb + ra) / (1 - ω * cc)
    W⁰ = im*ω⁰^3*im
    Z  = Z_Rancic(W⁰)
    z  = (Z^0.25)*2
    x, y = reim(z)

    kxy = abs(y) > abs(x)
    xx = x
    yy = y
    !kxy && ( x = 1 - abs(yy) )
    !kxy && ( y = 1 - abs(xx) )

    xf = x
    yf = y
    ( X < Y ) && ( xf = y  )
    ( X < Y ) && ( yf = x  )
    x = xf
    y = yf

    return x, y
end

W_Rancic(Z) = sum(A_Rancic[k] * Z^(k-1) for k in 1:length(A_Rancic))
Z_Rancic(W) = sum(B_Rancic[k] * W^(k-1) for k in 1:length(B_Rancic))
